{-# OPTIONS_GHC -fno-warn-orphans #-}
module Darcs.Patch.Prim
    ( PrimApply(..)
    , PrimCanonize(..)
    , PrimClassify(..)
    , PrimConstruct(..)
    , PrimDetails(..)
    , PrimMangleUnravelled(..)
    , PrimPatch
    , PrimRead(..)
    , PrimShow(..)
    , PrimSift(..)
    , Mangled
    , Unravelled
    ) where

import Darcs.Patch.Prim.Class
    ( PrimApply(..)
    , PrimCanonize(..)
    , PrimClassify(..)
    , PrimConstruct(..)
    , PrimDetails(..)
    , PrimMangleUnravelled(..)
    , PrimPatch
    , PrimRead(..)
    , PrimShow(..)
    , PrimSift(..)
    , Mangled
    , Unravelled
    )
